

/*

 Example showing how to handle hardware interrupts under DOS/4GW,
 whether the processor happens to be in real mode or protected mode
 at the time of the interrupt.

 Note that an external DPMI host (such as Windows) will reflect real
 mode interrupts to a protected mode handler rather than allowing the
 real mode handler to get them; installing a real mode handler in such
 an environment is not necessary.  However, it doesn't hurt to do so,
 so this program will handle interrupts as quickly as possible
 whether or not a DPMI host exists.

 The program is written to the DPMI (INT 31h) standard, so the code
 is the same for both environments.

 Written by Dan Teven, Rational Systems Inc., 3/26/92
 Last updated Feb. 8, 1994
*/

#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <stdlib.h>
#include <string.h>

#include "boulder.h"

const int KB_IRQ = 0x9;

#define	D32RealSeg(P)	((((DWORD) (P)) >> 4) & 0xFFFF)
#define	D32RealOff(P)	(((DWORD) (P)) & 0xF)

typedef unsigned short WORD;
typedef unsigned long DWORD;

typedef struct _dpmi_callregs
{
	DWORD edi, esi, ebp, esp;
	DWORD ebx, edx, ecx, eax;
	WORD flags;
	WORD es, ds;
	WORD fs, gs;
	WORD ip, cs;
	WORD sp, ss;
} DPMI_CALLREGS;

extern "C" void __interrupt pmhandler (void);
extern "C" void __interrupt __far rmhandler (void);
extern "C" unsigned short lowmem_seg;

WORD orig_rm_seg;
WORD orig_rm_off;
DWORD orig_pm_off;
WORD orig_pm_sel;
DPMI_CALLREGS dr;
char *lowp;

unsigned short lowmem_seg = 0;

// ######################################################################
// # void *D32DosMemAlloc (DWORD size, WORD *selector)
// ######################################################################
void *D32DosMemAlloc (DWORD size, WORD *selector)
{
   union REGS r;

   r.x.eax = 0x0100;                  //  DPMI allocate DOS memory
   r.x.ebx = (size + 15) >> 4;        // Number of paragraphs requested
   int386 (0x31, &r, &r);

   if (r.x.cflag) return ((DWORD) 0);                    //   Failed

   if (selector) *selector = r.w.dx;

   return (void *) ((r.x.eax & 0xFFFF) << 4);
}

// ######################################################################
// # void deinstall_kbd_handler()
// ######################################################################
void deinstall_kbd_handler()
{
  // Clean up.
  // ############

  union REGS r;
  r.x.eax = 0x0201;                 // DPMI set real mode vector
  r.x.ebx = (DWORD) KB_IRQ;
  r.x.ecx = (DWORD) orig_rm_seg;    // CX:DX == real mode &handler
  r.x.edx = (DWORD) orig_rm_off;
  int386 (0x31, &r, &r);

  r.x.eax = 0x0205;                 // DPMI set protected mode vector
  r.x.ebx = (DWORD) KB_IRQ;
  r.x.ecx = (DWORD) orig_pm_sel;
  r.x.edx = orig_pm_off;
  int386 (0x31, &r, &r);
}


// ######################################################################
// # void install_kbd_handler()
// ######################################################################
void install_kbd_handler()
{
  // Save the starting real-mode and protected-mode handler addresses.
  // #####################################################################

  union REGS r;
  void far *fh;
  r.x.eax = 0x0204;            // DPMI get protected mode vector
  r.h.bl = KB_IRQ;
  int386 (0x31, &r, &r);
  orig_pm_sel = (WORD) r.x.ecx;
  orig_pm_off = r.x.edx;

  r.x.eax = 0x0200;            // DPMI get real mode vector
  r.h.bl = KB_IRQ;
  int386 (0x31, &r, &r);
  orig_rm_seg = (WORD) r.x.ecx;
  orig_rm_off = (WORD) r.x.edx;

  // # Allocate 128 bytes of DOS memory for the real-mode handler,
  // #    which must of course be less than 128 bytes long, plus 256
  // #    bytes for the data.  Then copy the real-mode handler to DOS
  // #    memory and initialize the data to zero.
  // #####################################################################

  if (! (lowp = (char*) D32DosMemAlloc (128 + 4 + 256 + 1, &lowmem_seg)))
  {
     printf ("Couldn't get low memory!\n");
     exit (1);
  }
  _fmemcpy ((char __far *) lowp, (void __far *) rmhandler, 128);
  _fmemset ((char __far *) lowp + 128, 0, 4 + 256 + 1);
  fh = (void __far *) pmhandler;

  // # Install the new handlers.  The memory touched by the protected mode
  // #     handler needs to be locked, in case we are running under VMM
  // #     or an external DPMI host.
  // #####################################################################

  // for the Protected Mode Handler...
  // ####################################
  r.x.eax = 0x0205;                  // DPMI set protected mode vector
  r.x.ebx = (DWORD) KB_IRQ;
  r.x.ecx = (DWORD) FP_SEG(fh);
  r.x.edx = FP_OFF(fh);
  int386 (0x31, &r, &r);

  r.x.eax = 0x0600;                  // DPMI lock linear region
  r.x.ebx = ((DWORD) pmhandler) >> 16;
  r.x.ecx = ((DWORD) pmhandler) & 0xFFFF;
  r.x.esi = 0;
  r.x.edi = 256;                     // lock 256 bytes
  int386 (0x31, &r, &r);

  // for the Real Mode Handler...
  // ####################################
  r.x.eax = 0x0201;                // DPMI set real mode vector
  r.x.ebx = (DWORD) KB_IRQ;
  r.x.ecx = D32RealSeg(lowp);      // CX:DX == real mode &handler
  r.x.edx = D32RealOff(lowp);
  int386 (0x31, &r, &r);
  atexit(deinstall_kbd_handler);
}

// ######################################################################
// # short get_kb_command()
// ######################################################################
short get_kb_command()
{
    short wert = 0;
    if (*(lowp+130+1) ) wert |= esc;

    if (*(lowp+130+0x4d     ) ) wert |= right;
    if (*(lowp+130+0x4d+0x80) ) wert |= right;
    if (*(lowp+130+0x48     ) ) wert |= up;
    if (*(lowp+130+0x48+0x80) ) wert |= up;
    if (*(lowp+130+0x50     ) ) wert |= down;
    if (*(lowp+130+0x50+0x80) ) wert |= down;
    if (*(lowp+130+0x4b     ) ) wert |= left;
    if (*(lowp+130+0x4b+0x80) ) wert |= left;

    if (*(lowp+130+0x2a) ) wert |= shift;
    if (*(lowp+130+0x36) ) wert |= shift;

    if (*(lowp+130+28) || *(lowp+130+28 + 128)) wert |= enter;

    if ( (*(lowp+130+29) || *(lowp+130+29 + 128)) &&
         (*(lowp+130+56) || *(lowp+130+56 + 128)) &&
         (*(lowp+130+83) || *(lowp+130+83 + 128))
       )
       exit(-1);

    return wert;

// TODO: Pause + PrntScr fehlt
//       configure_kb()

}


// ######################################################################
// # void test_kbd_handler()
// ######################################################################
void test_kbd_handler()
{
  char *crt = (char *) 0xb8000L;
  union REGS r;
  struct SREGS sr;
  void far *fh;
  int i;

  puts ("Kick the keyboard.... press ESC to cancel\n");

  // # Wait for the ESC key to be pressed.  This loop has a good mix of
  // #       time spent in real mode and protected mode, so the upper left
  // #       hand corner of your color screen will toggle between 'R' and 'P'.
  // ########################################################################
  while (1)
  {
     if (*(lowp + 130 +1 )) break;           /* ESC pressed ? */
     for (i=0;i<256;i++)
         if ( *(lowp + 130 +i) ) crt[2*i] =  '*'; else crt[2*i] =  '.';
         // * auf Bildschirm malen

     // Explicitly go down to real mode to ask if a key is ready.
     // ###############################################################

     r.x.eax = 0x300;   // DPMI signal real mode interrupt
     r.x.ebx = 0x16;
     r.x.ecx = 0;
     fh = (void far *) &dr;
     _fmemset (fh, 0, sizeof(DPMI_CALLREGS));
     dr.eax = 0x0100;
     sr.ds = 0;
     sr.es = FP_SEG(fh);
     r.x.edi = FP_OFF(fh);
     int386x (0x31, &r, &r, &sr);
  }
}


#if 0
// ######################################################################
// # void main ()
// ######################################################################
void main ()
{

   install_kbd_handler();

   test_kbd_handler();

   deinstall_kbd_handler();

   puts ("Bye.\n");

}
#endif
